from realtrends import TrendsFetcher

import pandas

# Module to fetch approximate real/absolute search volumes for a specified term

class RealTrendsFetcher:
    def __get_reference_index(self, low_volume_profile):
        # search for the highest index, this will always be 100 as we are looking at this term alone
        reference_index = low_volume_profile.index[low_volume_profile[self.__ladder[0]] == 100]

        return reference_index

    # calculate what the maximum volume index represents as an absolute value
    # we assume that the lowest nonzero index represents a single search
    # so to get the absolute value at max, we divide the relative volume at max (100) by the relative search volume
    # at the nonzero minimum
    def __get_reference_abs(self, low_volume_profile):
        lowest_nonzero = low_volume_profile[low_volume_profile.gt(0)].min(0)
        return 100.0 / lowest_nonzero

    def __get_low_reference(self):
        relative_fetcher = TrendsFetcher()
        relative_fetcher.scrape_trend([self.__ladder[0]], geo=self.__geo, time_range=self.__time_range, tz=self.__tz)

        low_volume_profile = relative_fetcher.trends_data

        low_ref_index = self.__get_reference_index(low_volume_profile)
        low_abs_vol = self.__get_reference_abs(low_volume_profile)
        return low_ref_index, low_abs_vol

    # function to "climb" the ladder one step
    # gets a DataFrame with relative volumes for low_term and high_term
    # from this DataFrame, finds low_rel_vol at low_ref_index
    # high_ref_index is then selected to be where high_term is highest (always 100)
    # high_abs_vol is the absolute volume at this index, given by low_abs_vol * (high_rel_vol / low_rel_vol)
    # high_ref_index, high_abs_vol are then returned (these may then be passed to __step_absolute() again later to
    # generate the next index and volume
    def __step_absolute(self, low_term, high_term, low_ref_index, low_abs_vol):
        step_fetcher = TrendsFetcher()
        step_fetcher.scrape_trend([low_term, high_term],
                                  geo=self.__geo,
                                  time_range=self.__time_range,
                                  tz=self.__tz)
        step_data = step_fetcher.trends_data

        #print(step_data)

        # get relative value where low term popularity is highest
        low_rel_vol = int(step_data.loc[low_ref_index, low_term])

        # todo
        # for now just use index where value is 100 as reference point
        # in the future change this to not use very early data points (could
        # go out of range by the time the next request is made) or very
        # late points (where values are likely to change)
        high_ref_index = step_data.index[step_data[high_term] == 100]
        high_rel_vol = 100

        high_abs_vol = low_abs_vol * (high_rel_vol / low_rel_vol)
        # print(high_abs_vol)

        return high_ref_index, high_abs_vol

    def __transform_keyword_data(self, ladder_term, ladder_ref_index, ladder_abs_vol):
        # get relative data between ladder item and keyword
        keyword_fetcher = TrendsFetcher()
        keyword_fetcher.scrape_trend([ladder_term, self.__search_term],
                                     geo=self.__geo,
                                     time_range=self.__time_range,
                                     tz=self.__tz)

        search_term_data = keyword_fetcher.trends_data

        #print(search_term_data)

        # if ladder_term maximum relative volume is low this means it is not comparable to search_term, so we return
        # false, this then leads to the next ladder term having its reference index and absolute volume calculated
        # and passed, this is repeated until a comparable ladder term is found
        if search_term_data[ladder_term].max() < 10:
            return False

        # transform 2 column dataframe containing relative data into 1 column
        # dataframe containing absolute data for keyword

        ladder_rel_vol = int(search_term_data.loc[ladder_ref_index, ladder_term])

        scale_factor = ladder_abs_vol / ladder_rel_vol
        search_term_data[self.__search_term] = search_term_data[self.__search_term].apply(lambda x: round(x * scale_factor))
        search_term_data.drop(labels=[ladder_term], axis="columns", inplace=True)

        self.real_trends_data = search_term_data
        return True

    # member function to get real trends volume data
    def scrape_real(self, search_term, geo="", time_range="1-H",
                    tz="0", save_file="", retry=True,
                    ladder = ["englishcombe", "bathampton", "keynsham", "chippenham", "swindon", "bath", "london"]):
        self.__search_term = search_term
        self.__geo = geo
        self.__time_range = time_range
        self.__tz = tz

        self.__ladder = ladder

        # find a time and the absolute volume at this time using the first ladder item (default "englishcombe")
        low_ref_index, low_abs_vol = self.__get_low_reference()

        # check keyword against ladder base item (default "englishcombe")
        # if similar then no need to go up the ladder
        if self.__transform_keyword_data(self.__ladder[0], low_ref_index, low_abs_vol):
            return

        # Check keyword against ladder item 2, 3...
        i = 0
        while i < len(self.__ladder) - 1:
            comp1 = self.__ladder[i]
            comp2 = self.__ladder[i+1]

            # find
            high_ref_index, high_abs_vol = self.__step_absolute(comp1, comp2, low_ref_index, low_abs_vol)

            if self.__transform_keyword_data(comp2, high_ref_index, high_abs_vol):
                break

            low_ref_index = high_ref_index
            low_abs_vol = high_abs_vol
            i += 1

    def __init__(self):
        self.__search_term = ""
        self.__geo = ""
        self.__time_range = ""
        self.__tz = ""

        # set of values increasing ~exponentially in popularity
        self.__ladder = []
        # index in low volume data to use as reference point
        self.__low_ref_index = ""
        self.__low_abs_vol = 0

        # real volume data for external use
        self.real_trends_data = pandas.DataFrame()


import io
import re
import urllib.parse
from datetime import datetime, timezone

import pandas
import pycurl
from dateutil.relativedelta import relativedelta


class TrendsFetcher:
    __cookie = ""

    # Temporary headers taken from firefox on my machine
    __default_cookie_header = [
        "User-Agent: Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:78.0) Gecko/20100101 Firefox/78.0",
        "Accept: text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,*/*;q=0.8",
        "Accept-Language: en-GB,en;q=0.5",
        "Connection: keep-alive",
        "Upgrade-Insecure-Requests: 1",
        "Cache-Control: max-age=0"
    ]

    __default_token_header = [
        "User-Agent: Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:78.0) Gecko/20100101 Firefox/78.0",
        "Accept: application/json, text/plain, */*",
        "Accept-Language: en-GB,en;q=0.5",
        "Connection: keep-alive",
        "Referer: https://trends.google.com/trends/explore?q=snow&geo=US",
        "TE: Trailers"
    ]

    __default_csv_header = [
        "User-Agent: Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:78.0) Gecko/20100101 Firefox/78.0",
        "Accept: text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,*/*;q=0.8",
        "Accept-Language: en-GB,en;q=0.5",
        "Connection: keep-alive",
        "Referer: https://trends.google.com/trends/explore?q=snow&geo=US",
        "Upgrade-Insecure-Requests: 1",
        "TE: Trailers"
    ]

    # Fetch a cookie, this is needed to fetch a token
    def __get_cookie(self):
        cookie_curl = pycurl.Curl()

        cookie_curl.setopt(pycurl.URL,
                           'https://trends.google.com/trends/explore?geo=US&q=snow')
        cookie_curl.setopt(pycurl.HTTPHEADER, self.__default_cookie_header)

        byte_data = io.BytesIO()
        cookie_curl.setopt(pycurl.WRITEHEADER, byte_data)

        cookie_curl.perform_rb()
        cookie_curl.close()
        # convert header into a string and extract cookie
        header = byte_data.getvalue().decode("utf8")

        header = header.split("Set-Cookie:")[1]
        cookie = header.split(";")[0] + ";"

        TrendsFetcher.__cookie = cookie

    def __generate_token_query_request_comparison_item_list(self):
        time_phrase = {
            "1-H": "now 1-H", "4-H": "now 4-H", "1-d": "now 1-d",
            "7-d": "now 7-d", "1-m": "today 1-m", "3-m": "today 3-m",
            "12-m": "today 12-m", "5-y": "today 5-y",
        }
        token_time = time_phrase.get(self.__time_range)
        comparison_item_list = ""
        for kw in self.__keywords:
            comparison_item_list_element = """
            "keyword":"%s","geo":"%s","time":"%s"
            """ % (kw, self.__geo, token_time)
            comparison_item_list_element = "{" + comparison_item_list_element + "},"
            comparison_item_list += comparison_item_list_element
        # Return the list except the last character, which is an unneeded comma
        return comparison_item_list[:-1]

    def __generate_token_query_request_comparisonItem(self):
        comparison_item = self.__generate_token_query_request_comparison_item_list()
        comparison_item = """\"comparisonItem":[""" + comparison_item + """],"""
        return comparison_item

    def __generate_token_query_request(self):
        token_query_request = \
            self.__generate_token_query_request_comparisonItem() + """\"category":0,"property":"\""""
        token_query_request = "{" + token_query_request + "}"
        return token_query_request

    # Fetch a token for a google trends query, this is needed to get csv data on the query
    def __get_token(self):
        token_curl = pycurl.Curl()
        token_address = "https://trends.google.com/trends/api/explore?"
        token_query = urllib.parse.urlencode(
            {"hl": "en-US", "tz": self.__tz,
             "req": self.__generate_token_query_request()})

        token_url = token_address + token_query
        token_curl.setopt(pycurl.URL, token_url)
        token_curl.setopt(pycurl.HTTPHEADER, self.__default_token_header)
        # get body as string
        response = token_curl.perform_rs()
        token_curl.close()
        # use first token in response
        self.__token = response.split("\"token\":\"")[1]
        self.__token = self.__token.split("\",")[0]

    # generates the csv query request string time component
    def __generate_csv_query_request_time(self):
        resolution = {
            "1-H": "MINUTE", "4-H": "MINUTE", "1-d": "EIGHT_MINUTE",
            "7-d": "HOUR", "1-m": "DAY", "3-m": "DAY",
            "12-m": "WEEK", "5-y": "WEEK"
        }
        interval_unit = {
            "H": "hours", "d": "days", "m": "months", "y": "years"
        }
        interval_keyword = interval_unit.get(self.__time_range.split("-")[1])
        interval_quantity = int(self.__time_range.split("-")[0])
        start_date_time = \
            datetime.now(timezone.utc) \
            - relativedelta(
                **{interval_keyword: interval_quantity}
                            )

        if resolution.get(self.__time_range) in {"MINUTE", "EIGHT_MINUTE", "HOUR"}:
            start_date_time = start_date_time.strftime("%Y-%m-%dT%H\\\\:%M\\\\:%S")
            end_date_time = datetime.now(timezone.utc).strftime("%Y-%m-%dT%H\\\\:%M\\\\:%S")
        else:
            start_date_time = start_date_time.strftime("%Y-%m-%d")
            end_date_time = datetime.now(timezone.utc).strftime("%Y-%m-%d")

        return """
        "time":"%s %s",
        "resolution":"%s"
        """ % (start_date_time, end_date_time, resolution.get(self.__time_range))

    def __generate_csv_query_request_comparison_item_list_geo(self):
        if self.__geo != "":
            return """\"country":"%s\"""" % self.__geo
        else:
            return ""

    def __generate_csv_query_request_comparison_item_list(self):
        comparison_item_list = ""
        for kw in self.__keywords:
            comparison_item_list += """
            {
                "geo":{%s},
                "complexKeywordsRestriction":{
                    "keyword":[
                        {
                            "type":"BROAD",
                            "value":"%s"
                        }
                    ]
                }
            },
            """ % (self.__generate_csv_query_request_comparison_item_list_geo(), kw)
        # Last character is unneeded comma, slice:
        return "[%s]" % comparison_item_list[:-1]

    def __generate_csv_query_request_request_options(self):
        backends = {"1-H": "CM", "4-H": "CM", "1-d": "CM",
                    "7-d": "CM", "1-m": "IZG", "3-m": "IZG",
                    "12-m": "IZG", "5-y": "IZG"}
        select_backend = backends.get(self.__time_range)
        return """{"property":"","backend":"%s","category":0}""" % select_backend

    def __generate_csv_query_request(self):
        csv_query_request = """
        {%s,"locale":"en-US","comparisonItem":%s,\"requestOptions":%s}
        """ % (self.__generate_csv_query_request_time(),
               self.__generate_csv_query_request_comparison_item_list(),
               self.__generate_csv_query_request_request_options())
        return csv_query_request

    def __get_csv(self):
        csv_address = """https://trends.google.com/trends/api/widgetdata/multiline/csv?"""
        csv_query = urllib.parse.urlencode(
            {"req": self.__generate_csv_query_request(),
             "token": self.__token, "tz": self.__tz})

        request_curl = pycurl.Curl()
        request_curl.setopt(pycurl.URL, csv_address + csv_query)
        request_curl.setopt(pycurl.HTTPHEADER, self.__default_csv_header)

        response_string = request_curl.perform_rs()
        request_curl.close()
        # Delete top two lines of response, ready for conversion to DataFrame
        response_string = re.sub(".*?\n\n", "", response_string)
        # Put response into buffer DataFrame
        response_io_string = io.StringIO(response_string)
        self.__trends_data_buffer = pandas.read_csv(response_io_string, sep=",")

    def scrape_trend(
            self, keywords, geo="", time_range="12-m",
            tz="0", save_file="", retry=True, simple_indexes=True):
        self.__keywords = keywords
        self.__geo = geo
        self.__time_range = time_range
        self.__tz = tz

        self.__get_token()

        fetch_fail = True
        retry_csv = 0
        retry_token = 0

        # First retry fetching CSV, if this fails again, retry fetching token:
        while fetch_fail:
            self.__get_csv()
            if retry == False:
                break
            if self.__trends_data_buffer.empty:
                fetch_fail = True
            else:
                fetch_fail = False
            if retry_csv >= 1:
                retry_csv = -1
                retry_token += 1
                # print("get_token retry %s" % retry_token)
                self.__get_token()
            retry_csv += 1
            # print("get_csv retry %s" % retry_csv)

        # make the first index "date_time" and the following columns simply
        # their keywords. Replace "<1" with "1", convert all entries to
        # integers
        if simple_indexes:
            # make first column name simply "date_time"
            self.__trends_data_buffer.rename(
                columns={self.__trends_data_buffer.columns[0]: "date_time"},
                inplace=True)
            # make column names just the keyword
            for n in self.__trends_data_buffer.columns:
                self.__trends_data_buffer.rename(columns={n: n.split(":")[0]},
                                                 inplace=True)
            # set "date_time" column as index
            self.__trends_data_buffer.set_index("date_time", inplace=True)

            # sanitise for "<1"
            for i in self.__trends_data_buffer.columns:
                self.__trends_data_buffer[i] = [str(x).split("<")[-1] if str(x).find("<") != -1 else str(x) for x in self.__trends_data_buffer[i]]
            #convert al values back to integers
            for i in self.__trends_data_buffer.columns:
                self.__trends_data_buffer[i] = [int(x) for x in self.__trends_data_buffer[i]]

        self.trends_data = self.__trends_data_buffer.copy()

        if save_file != "":
            self.trends_data.to_csv(save_file)

    def __init__(self):
        self.__keywords = ""
        self.__geo = ""
        self.__time_range = ""
        self.__tz = ""

        self.trends_data = pandas.DataFrame()

        if TrendsFetcher.__cookie == "":
            self.__get_cookie()
        self.__default_token_header.append("Cookie: " + self.__cookie)


from setuptools import setup

setup(name='realtrends',
      version='1.1.0',
      description='Package to scrape google trends for absolute/real and relative search volumes. Data is presented as '
                  'a DataFrame.',
      url='https://github.com/Litharge/realtrends',
      author='Robert Chambers',
      author_email='1robertchambers@gmail.com',
      license='MIT',
      packages=['realtrends'],
      install_requires=[
            'python-dateutil',
            'pandas',
            'pycurl'
      ],
      zip_safe=False)